#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <vector>

/* for srand() */
#include <cstdlib>
#include <ctime>

using namespace std;

#include "bt_peer.h"
#include "event.h"
#include "topology.h"
#include "util.h"
/* added by Ruchir for Poisson distributed Random Numbers*/
#include <stdlib.h>

/**********
 * MACROS *
 **********/




/*************
 * CONSTANTS *
 *************/


const unsigned int MAX_EVENTS_PER_PEER = 5;
const char *DEFAULT_TOPO_FILE = "topology.in";


/*********** 
 * GLOBALS * 
 ***********/


unsigned int g_tick = 0;
int done = 0;
unsigned long long g_peerslogged = 0;
unsigned long long g_peersidle = 0;


/***********************
 * FUNCTION PROTOTYPES *
 ***********************/

/**
 * Parses the command line arguments and processes them.  This includes 
 * loading the topology file and initializing the topology and events.
 * @param argc		the argument count
 * @param argv		the argument vector
 * @return 0 on success, non-zero on error.
 */
static int init_sim(int argc, char *argv[]);

/**
 * Prints the command line arguments usage.
 * @param argv0		the program name argument
 */
static void print_usage(const char *argv0);

/**
 * Join event handler.  Starts up a bt_peer.
 * @param peer		the peer to start up
 */
static __inline__ void join_handler(bt_peer *peer);

/**
 * Leave event handler.  Instructs a bt_peer to leave the network.
 * @param peer		the peer to tell to leave
 */
static __inline__ void leave_handler(bt_peer *peer);

/**
 * Choke timer event handler.  Instructs the peer to choke.
 * @param peer		the peer to tell to choke
 */
static __inline__ void choke_handler(bt_peer *peer);

/**
 * Rerequest timer event handler.  Instructs the peer to rerequest peers
 * from tracker.
 * @param peer		the peer to tell to rerequest
 */
static __inline__ void rerequest_handler(bt_peer *peer);

/**
 * Data transfer timer event handler.  Tells the topology to calculate 
 * data transfer progress.
 */
static __inline__ void transfer_handler();

/**
 * Logging timer event handler.  Iterates through bt_peer's and instructs 
 * them to log their data.
 */
static __inline__ void logging_handler();

/**
 * Adds all the join events to the event queue.
 */
static void init_joins();

/**
 * Logs messages at the end of the simulation.
 */
static __inline__ void log_report();

/************************
 * FUNCTION DEFINITIONS *
 ************************/

int main(int argc, char *argv[])
{
	if (init_sim(argc, argv)) {
		return 1;
	}

	/* main event loop */
	done = 0;

	struct event *ev;
	while (!done && (ev = event_pop()) != NULL) {
		g_tick = ev->fire_time;
		switch (ev->type) {
			case E_JOIN:
				join_handler(ev->peer);
				break;
			case E_LEAVE:
				leave_handler(ev->peer);
				break;
			case E_TRANSFERBLOCKS:
				assert(!ev->peer);
				transfer_handler();
				break;
			case E_OPTUNCHOKE:
				choke_handler(ev->peer);
				break;
			case E_REREQUEST:
				rerequest_handler(ev->peer);
				break;
			case E_LOGGING:
				assert(!ev->peer);
				logging_handler();
				break;
			default:
				assert(0);
				break;
		}
		event_free(ev);
	}

	log_report();

	return 0;
}

static int init_sim(int argc, char *argv[])
{
	int ret = 0;
	const char *topo_file = DEFAULT_TOPO_FILE;
	int c;
	struct event *ev = NULL;

	while ((c = getopt(argc, argv, "hdt:l:")) != EOF) {
		switch (c) {
			case 'h':
				print_usage(argv[0]);
				return 1;
			case 'd':
				debugging_enable();
				break;
			case 't':
				topo_file = optarg;
				break;
			case 'l':
				if (open_logfile(optarg)) {
					return 1;
				}
				logging_enable();
				break;
			default:
				print_usage(argv[0]);
				return 1;
		}
	}

	EXPECT_ZERO(init_topology(topo_file));

	g_tick = 0;

	init_events(get_num_peers() * MAX_EVENTS_PER_PEER);
	init_joins();

	ev = event_alloc();
	ev->fire_time = TICKS_PER_SECOND * LOGEVENT_MS / 1000;
	ev->type = E_LOGGING;
	ev->peer = NULL;

	event_insert(ev);

	ev = event_alloc();
	ev->fire_time = 0;
	ev->type = E_TRANSFERBLOCKS;
	ev->peer = NULL;

	event_insert(ev);

	/* seed the random number generator*/
	/* added by Ruchir on Jun 10'05*/
	srand((unsigned)time(0));
	



_cleanup:
	if (ret) {
		fprintf(stderr, "Failed to load topology file.\n");
	}
	return ret;
}

static void print_usage(const char *argv0)
{
	fprintf(stderr, "Usage: %s [-h] [-d] [-l logfile] [-t topology]\n",
			argv0);
	fprintf(stderr, "\tDefaults: topology=%s\n",
			"\t          debugging=false\n",
			"\t	     logging=false\n",
			DEFAULT_TOPO_FILE);
}

static void init_joins()
{
	bt_peer *peers = get_peers();
	assert(peers);

	/* TODO: allow configuration of arrival process */
	int random_joins = get_arrivals_are_random();


	assert(!random_joins);

	/* order of joins => AS 1 -> 2 -> 3 ..... -> 14 -> 1 -> 2 .......*/
	unsigned int num_univ_nodes = get_num_univ_nodes();
	unsigned int as_nodes = get_num_peers() -  num_univ_nodes;

	/* first finish off with all the univ nodes 
	for (unsigned int i = 0; i < num_univ_nodes; i++) {
		struct event *ev = event_alloc();
		assert(ev);
		ev->fire_time = g_tick;
		ev->peer = &peers[i];
		ev->type = E_JOIN;
		event_insert(ev);

	}*/


	/* TODO: change these two values to be non-static*/
	unsigned int num_per_as = 50;
	unsigned int num_as = 14;	/* for university nodes */

	/*assert((num_per_as * num_as ) == as_nodes);*/

	for(unsigned int i = 0; i < num_per_as ; i++)
	{
		for (unsigned int j = 0; j < num_as; j++) 
		{
			struct event *ev = event_alloc();
			assert(ev);
			ev->fire_time = g_tick;
			ev->peer = &peers[j*num_per_as + i];
			ev->type = E_JOIN;
			event_insert(ev);
		}
	}
}

static __inline__ void join_handler(bt_peer *peer)
{
	static int first = 1;
	assert(peer);

	bt_peer *peers = get_peers();
	assert(peers);
	unsigned int peer_id = peer - peers;

	if (first && ((!get_seed_is_cable() && peer_id < get_num_univ_nodes())
			|| (get_seed_is_cable() && peer_id >=
			get_num_univ_nodes()))) {
		peer->join(true, peer_id+1, get_num_pieces());
		first = 0;
	} else {
		peer->join(false, peer_id+1, get_num_pieces());
	}
}

static __inline__ void leave_handler(bt_peer *peer)
{
	assert(peer);

	bt_peer *peers = get_peers();
	LOG_MSG("Peer[%d] is leaving the network.\n", peer-peers);

	peer->leave();
}

static __inline__ void choke_handler(bt_peer *peer)
{
	assert(peer);

	peer->choker();
}

static __inline__ void rerequest_handler(bt_peer *peer)
{
	assert(peer);

	peer->rerequest();
}

static __inline__ void transfer_handler()
{
	calc_data_transfers(g_tick);
}

static __inline__ void logging_handler()
{
	bt_peer *peers = get_peers();
	assert(peers);
	bt_peer *peer;

	unsigned int num_peers = get_num_peers();
	unsigned int num_idle = 0;
	unsigned int num_done = 0;
	for (unsigned int i = 0; i < num_peers; i++) {
		peer = &peers[i];
		if (peer->done) {
			num_done++;
		} else {
			g_peerslogged++;
		       	if (peer->peer_is_idle()) {
				num_idle++;
			}
		}
		peer->log_peer_info();
	}

	LOG_MSG("%u peers.  %u idle, (%f%% idle).  %u completed.\n",
			num_peers,
			num_idle,
			100 * (float)num_idle / (float)num_peers,
			num_done);

	g_peersidle += num_idle;

	if (num_done == num_peers) {
		done = 1;
	}

	struct event *ev = event_alloc();
	ev->fire_time = g_tick + TICKS_PER_SECOND * LOGEVENT_MS / 1000;
	ev->type = E_LOGGING;
	ev->peer = NULL;
	event_insert(ev);
}

static __inline__ void log_report()
{
	log_bw_usage();
	LOG_MSG("Total log peer product events: %llu.\n", g_peerslogged);
	LOG_MSG("Total idle events: %llu.\n", g_peersidle);
	LOG_MSG("Average %% of time a peer was idle: %f.\n",
			((double)g_peersidle / double(g_peerslogged)) * 100);
	unsigned long long total_tx = 0;
	for (unsigned int i = 0; i < DEGREES_OF_CLOSENESS; i++) {
		total_tx += g_closeness[i];
	}
	LOG_MSG("Percentage intra-headend data transfers: %f.\n",
			(double)g_closeness[INTRA_HEADEND] / (double)total_tx);
	LOG_MSG("Percentage intra-univ data transfers: %f.\n",
			(double)g_closeness[INTRA_UNIV] / (double)total_tx);
	LOG_MSG("Percentage intra-as data transfers: %f.\n",
			(double)g_closeness[INTRA_AS] / (double)total_tx);
	LOG_MSG("Percentage inter-* data transfers: %f.\n",
			(double)g_closeness[INTER] / (double)total_tx);

}
